# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This subpackage defines region shapes.
"""
from .annulus import *  # noqa
from .circle import *  # noqa
from .ellipse import *  # noqa
from .line import *  # noqa
from .point import *  # noqa
from .polygon import *  # noqa
from .rectangle import *  # noqa
from .text import *  # noqa
