# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This subpackage provides tools for reading and writing region files.
"""
from .core import *  # noqa

from .ds9.connect import *  # noqa
from .ds9.core import *  # noqa
from .ds9.read import *  # noqa
from .ds9.write import *  # noqa

from .crtf.connect import *  # noqa
from .crtf.core import *  # noqa
from .crtf.read import *  # noqa
from .crtf.write import *  # noqa

from .fits.connect import *  # noqa
from .fits.core import *  # noqa
from .fits.read import *  # noqa
from .fits.write import *  # noqa
