# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This subpackage provides the core functionality for regions.
"""

from .attributes import *  # noqa
from .bounding_box import *  # noqa
from .compound import *  # noqa
from .core import *  # noqa
from .mask import *  # noqa
from .metadata import *  # noqa
from .pixcoord import *  # noqa
from .regions import *  # noqa
