# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Geometry subpackage for low-level geometry functions.
"""

from .circular_overlap import *  # noqa
from .elliptical_overlap import *  # noqa
from .rectangular_overlap import *  # noqa
from .polygonal_overlap import *  # noqa
